//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#ifndef __ARM_OpenCL_Buffer_H_
#define __ARM_OpenCL_Buffer_H_

#include <CL/cl.h>
#include <iostream>

#include "ArmCLContext.h"

class ArmCLBuffer {
  public:
    ArmCLBuffer(cl_mem buffer,
            cl_mem_flags flags,
            size_t size,
            void* host_pr,
            cl_mem cpu_buffer,
            ArmCLContext* armcl_context);
    ~ArmCLBuffer();

    cl_mem buffer() const { return buffer_; }
    cl_mem cpu_buffer() const { return cpu_buffer_; }

    cl_mem_flags flags() const { return flags_; }
    size_t size() const { return size_; }
    void* host_ptr() const { return host_ptr_; }

    ArmCLContext* armcl_context() const { return armcl_context_; }

  private:
    void PrepareOpenCLCalls();

    const cl_mem buffer_;
    const cl_mem cpu_buffer_;

    cl_mem_flags flags_;
    size_t size_;
    void* host_ptr_;

    cl_context context_;
    cl_context cpu_context_;

    ArmCLContext* armcl_context_;

    cl_mem CL_API_CALL (*clCreateBuffer_p_) (cl_context context,
            cl_mem_flags flags,
            size_t size,
            void* host_ptr,
            cl_int* errcode_ret);
};

#endif
