//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLCommandQueue.hpp"
#include <sys/time.h>
#include <string.h>
#include <algorithm>
#include <atomic>
#include "/home/jiyeon/libsvm/svm.h"
#include "Tree.h"

static long long get_time_stamp(void)
{
    static __thread struct timeval tv;
    gettimeofday(&tv, NULL);
    return 1000000LL * tv.tv_sec + tv.tv_usec;
}

//ArmCLWorkGroupPool::ArmCLWorkGroupPool(int num_cus, cl_device_type device_type)
//    : num_cus_(num_cus), device_type_(device_type)
//{
//    num_ranges_ = 0;
//    wg_total_ = 0;
//    is_done_ = false;
//
//    pthread_mutex_init(&mutex_, NULL);
//}
//
//ArmCLWorkGroupPool::~ArmCLWorkGroupPool()
//{
//    pthread_mutex_destroy(&mutex_);
//}
//
//wgrange* ArmCLWorkGroupPool::next_work()
//{
//    this->lock();
//
//    if (wg_total_ == 0) {
//        this->unlock();
//        return NULL;
//    }
//
//    wgrange* return_wg_range = (wgrange*)malloc(sizeof(wgrange));
//    std::map<int, wgrange*>::iterator first_wg_range_it = wg_ranges_.begin();
//
//    wgrange* first_wg_range = first_wg_range_it->second;
//    if (chunk_size_ < first_wg_range->nwg) {
//        size_t wg_chunk = chunk_size_;
//
//        return_wg_range->start = first_wg_range->start;
//        return_wg_range->end = first_wg_range->start + wg_chunk;
//        return_wg_range->nwg = wg_chunk;
//
//        first_wg_range->start += wg_chunk;
//        first_wg_range->nwg -= wg_chunk;
//
//        wg_total_ -= wg_chunk;
//    } else {
//        size_t wg_chunk = first_wg_range->nwg;
//
//        return_wg_range->start = first_wg_range->start;
//        return_wg_range->end = first_wg_range->end;
//        return_wg_range->nwg = first_wg_range->nwg;
//
//        free(first_wg_range);
//        wg_ranges_.erase(first_wg_range_it);
//
//        wg_total_ -= wg_chunk;
//        num_ranges_ -= 1;
//    }
//
//    this->unlock();
//    return return_wg_range;
//}
//
//wgrange* ArmCLWorkGroupPool::pop_work()
//{
//    if (wg_total_ == 0) return NULL;
//
//    size_t dynamic_chunk = wg_total_ / 10;
//    wgrange* return_wg_range = (wgrange*)malloc(sizeof(wgrange));
//
//    if (device_type_ == CL_DEVICE_TYPE_CPU) {
//        std::map<int, wgrange*>::iterator first_wg_range_it = wg_ranges_.begin();
//        wgrange* first_wg_range = first_wg_range_it->second;
//        if (dynamic_chunk < first_wg_range->nwg) {
//            size_t wg_chunk = dynamic_chunk;
//
//            return_wg_range->end = first_wg_range->start + wg_chunk;
//            return_wg_range->start = first_wg_range->start;
//            return_wg_range->nwg = wg_chunk;
//
//            first_wg_range->start += wg_chunk;
//            first_wg_range->nwg -= wg_chunk;
//
//            wg_total_ -= wg_chunk;
//        } else {
//            size_t wg_chunk = first_wg_range->nwg;
//
//            return_wg_range->start = first_wg_range->start;
//            return_wg_range->end = first_wg_range->end;
//            return_wg_range->nwg = first_wg_range->nwg;
//
//            free(first_wg_range);
//            wg_ranges_.erase(first_wg_range_it);
//
//            wg_total_ -= wg_chunk;
//            num_ranges_ -= 1;
//        }
//    } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//        std::map<int, wgrange*>::iterator last_wg_range_it = wg_ranges_.begin();
//        for (int i = 0; i < wg_ranges_.size()-1; i++)
//            last_wg_range_it++;
//
//        wgrange* last_wg_range = last_wg_range_it->second;
//        if (dynamic_chunk < last_wg_range->nwg) {
//            size_t wg_chunk = dynamic_chunk;
//
//            return_wg_range->end = last_wg_range->end;
//            return_wg_range->start = last_wg_range->end - wg_chunk;
//            return_wg_range->nwg = wg_chunk;
//
//            last_wg_range->end -= wg_chunk;
//            last_wg_range->nwg -= wg_chunk;
//
//            wg_total_ -= wg_chunk;
//        } else {
//            size_t wg_chunk = last_wg_range->nwg;
//
//            return_wg_range->start = last_wg_range->start;
//            return_wg_range->end = last_wg_range->end;
//            return_wg_range->nwg = last_wg_range->nwg;
//
//            free(last_wg_range);
//            wg_ranges_.erase(last_wg_range_it);
//
//            wg_total_ -= wg_chunk;
//            num_ranges_ -= 1;
//        }
//    }
//
//    return return_wg_range;
//}
//
//wgrange* ArmCLWorkGroupPool::pop_work(size_t num_wgs)
//{
//    if (wg_total_ == 0) return NULL;
//
//    wgrange* return_wg_range = (wgrange*)malloc(sizeof(wgrange));
//
//    if (device_type_ == CL_DEVICE_TYPE_CPU) {
//        std::map<int, wgrange*>::iterator first_wg_range_it = wg_ranges_.begin();
//        wgrange* first_wg_range = first_wg_range_it->second;
//        if (chunk_size_ < first_wg_range->nwg) {
//            size_t wg_chunk = chunk_size_;
//
//            return_wg_range->end = first_wg_range->start + wg_chunk;
//            return_wg_range->start = first_wg_range->start;
//            return_wg_range->nwg = wg_chunk;
//
//            first_wg_range->start += wg_chunk;
//            first_wg_range->nwg -= wg_chunk;
//
//            wg_total_ -= wg_chunk;
//        } else {
//            return NULL;
//            //size_t wg_chunk = first_wg_range->nwg;
//
//            //return_wg_range->start = first_wg_range->start;
//            //return_wg_range->end = first_wg_range->end;
//            //return_wg_range->nwg = first_wg_range->nwg;
//
//            //free(first_wg_range);
//            //wg_ranges_.erase(first_wg_range_it);
//
//            //wg_total_ -= wg_chunk;
//            //num_ranges_ -= 1;
//        }
//    } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//        std::map<int, wgrange*>::iterator last_wg_range_it = wg_ranges_.begin();
//        for (int i = 0; i < wg_ranges_.size()-1; i++)
//            last_wg_range_it++;
//
//        wgrange* last_wg_range = last_wg_range_it->second;
//        if (chunk_size_ < last_wg_range->nwg) {
//            size_t wg_chunk = chunk_size_;
//
//            return_wg_range->end = last_wg_range->end;
//            return_wg_range->start = last_wg_range->end - wg_chunk;
//            return_wg_range->nwg = wg_chunk;
//
//            last_wg_range->end -= wg_chunk;
//            last_wg_range->nwg -= wg_chunk;
//
//            wg_total_ -= wg_chunk;
//        } else {
//            return NULL;
//            //size_t wg_chunk = last_wg_range->nwg;
//
//            //return_wg_range->start = last_wg_range->start;
//            //return_wg_range->end = last_wg_range->end;
//            //return_wg_range->nwg = last_wg_range->nwg;
//
//            //free(last_wg_range);
//            //wg_ranges_.erase(last_wg_range_it);
//
//            //wg_total_ -= wg_chunk;
//            //num_ranges_ -= 1;
//        }
//    }
//
//    return return_wg_range;
//}
//
//bool ArmCLWorkGroupPool::work_stealing(ArmCLWorkGroupPool* other_wg_pool)
//{
//    /* designate locking order with their node ids to avoid deadlock */
//    if (device_type_ == CL_DEVICE_TYPE_CPU) {
//        this->lock();
//        other_wg_pool->lock();
//    } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//        other_wg_pool->lock();
//        this->lock();
//    }
//
//    wgrange* wg_range = other_wg_pool->pop_work(num_cus_ * 4);
//    if (wg_range != NULL) {
//        if (device_type_ == CL_DEVICE_TYPE_CPU) {
//            other_wg_pool->unlock();
//            this->unlock();
//        } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//            this->unlock();
//            other_wg_pool->unlock();
//        }
//
//        this->push_work(wg_range);
//        return true;
//    }
//
//    if (device_type_ == CL_DEVICE_TYPE_CPU) {
//        other_wg_pool->unlock();
//        this->unlock();
//    } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//        this->unlock();
//        other_wg_pool->unlock();
//    }
//}
//
//void ArmCLWorkGroupPool::push_work(wgrange* wg_range)
//{
//    wg_ranges_.insert(std::pair<int, wgrange*>
//            (wg_range->start, wg_range));
//
//    wg_total_ += wg_range->nwg;
//    num_ranges_ += 1;
//}
//
//void ArmCLWorkGroupPool::coalesce_ranges()
//{
//}
//
//ArmCLWorkGroupScheduler::ArmCLWorkGroupScheduler(size_t num_available_cus,
//        ArmCLWorkGroupPool* cpu_wg_pool,
//        ArmCLWorkGroupPool* gpu_wg_pool,
//        cl_device_type device_type)
//    : num_available_cus_(num_available_cus), cpu_wg_pool_(cpu_wg_pool),
//    gpu_wg_pool_(gpu_wg_pool), device_type_(device_type)
//{
//    PrepareOpenCLCalls();
//
//    pthread_mutex_init(&mutex_schedule_queue_, NULL);
//    pthread_cond_init(&cond_schedule_queue_, NULL);
//
//    schedule_to_do_ = false;
//    thread_running_ = true;
//
//    pthread_create(&thread_, NULL, &ArmCLWorkGroupScheduler::ThreadFunc, this);
//}
//
//ArmCLWorkGroupScheduler::~ArmCLWorkGroupScheduler()
//{
//}
//
//bool ArmCLWorkGroupScheduler::IsIdle() const
//{
//    return !schedule_to_do_;
//}
//
//void ArmCLWorkGroupScheduler::Join()
//{
//    pthread_mutex_lock(&mutex_schedule_queue_);
//    if (schedule_to_do_)
//        pthread_cond_wait(&cond_schedule_queue_, &mutex_schedule_queue_);
//    pthread_mutex_unlock(&mutex_schedule_queue_);
//}
//
//void ArmCLWorkGroupScheduler::Launch(ArmCLWorkGroupAssignment* wga)
//{
//    pthread_mutex_lock(&mutex_schedule_queue_);
//    wga_ = wga;
//    schedule_to_do_ = true;
//    pthread_cond_signal(&cond_schedule_queue_);
//    pthread_mutex_unlock(&mutex_schedule_queue_);
//}
//
//void ArmCLWorkGroupScheduler::Run()
//{
//    ArmCLWorkGroupPool* my_wg_pool = NULL;
//    ArmCLWorkGroupPool* other_wg_pool = NULL;
//
//    if (device_type_ == CL_DEVICE_TYPE_CPU) {
//        my_wg_pool = cpu_wg_pool_;
//        other_wg_pool = gpu_wg_pool_;
//    } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//        my_wg_pool = gpu_wg_pool_;
//        other_wg_pool = cpu_wg_pool_;
//    }
//
//    while (thread_running_) {
//        pthread_mutex_lock(&mutex_schedule_queue_);
//        if (!schedule_to_do_)
//            pthread_cond_wait(&cond_schedule_queue_, &mutex_schedule_queue_);
//        pthread_mutex_unlock(&mutex_schedule_queue_);
//
//        struct timespec t1, t2;
//        clock_gettime(CLOCK_REALTIME, &t1);
//
//        wgrange* wg_range = NULL;
//
//        while (true) {
//            wgrange* wg_range = my_wg_pool->next_work();
//            if (wg_range == NULL) {
//                if (my_wg_pool->work_stealing(other_wg_pool)) {
//                    wg_range = my_wg_pool->next_work();
//                    if (wg_range == NULL) break;
//                } else {
//                    break;
//                }
//            }
//
//            //if (device_type_ == CL_DEVICE_TYPE_CPU) {
//            //    std::cout << "[CPU]: start: " << wg_range->start
//            //        << " end: " << wg_range->end
//            //        << " nwg: " << wg_range->nwg << std::endl;
//
//            //} else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//            //    std::cout << "[GPU]: start: " << wg_range->start
//            //        << " end: " << wg_range->end
//            //        << " nwg: " << wg_range->nwg << std::endl;
//            //}
//
//            size_t global_work_offset[1];
//            global_work_offset[0] = wg_range->start * wga_->local_work_size;
//
//            size_t global_work_size[1];
//            global_work_size[0] = wg_range->nwg * wga_->local_work_size;
//
//            size_t local_work_size[1];
//            local_work_size[0] = wga_->local_work_size;
//
//            //std::cout << " global_work_offset: " <<
//            //    global_work_offset[0]
//            //    << " global_work_size: " << global_work_size[0]
//            //    << " local_work_size: " << local_work_size[0]
//            //    << std::endl;
//
//            cl_int err;
//            err = clEnqueueNDRangeKernel_p_(wga_->command_queue,
//                    wga_->kernel,
//                    wga_->work_dim,
//                    global_work_offset,
//                    global_work_size,
//                    local_work_size,
//                    wga_->num_events_in_wait_list,
//                    wga_->event_wait_list,
//                    wga_->event);
//            if (err != CL_SUCCESS) {
//                std::cout << "ArmCL: work-group scheduler: "
//                    << " clEnqueueNDRangeKernel_p_ failed."
//                    << std::endl;
//            }
//
//            /* launch wg_range */
//            err = clFinish(wga_->command_queue);
//        }
//
//        free(wg_range);
//
//        clock_gettime(CLOCK_REALTIME, &t2);
//        double elapsed_time_us = ((t2.tv_sec - t1.tv_sec)*1000000000 + (t2.tv_nsec - t1.tv_nsec))/1000;
//        if (device_type_ == CL_DEVICE_TYPE_CPU) {
//            std::cout << "[CPU] elapsed_time: " << elapsed_time_us << std::endl;
//        } else if (device_type_ == CL_DEVICE_TYPE_GPU) {
//            std::cout << "[GPU] elapsed_time: " << elapsed_time_us << std::endl;
//        }
//
//        pthread_mutex_lock(&mutex_schedule_queue_);
//        schedule_to_do_ = false;
//        pthread_cond_signal(&cond_schedule_queue_);
//        pthread_mutex_unlock(&mutex_schedule_queue_);
//    }
//
//    return;
//}
//
//void ArmCLWorkGroupScheduler::PrepareOpenCLCalls()
//{
//    char* error;
//
//    clEnqueueNDRangeKernel_p_ = (cl_int (*) (cl_command_queue command_queue,
//                cl_kernel kernel,
//                cl_uint work_dim,
//                const size_t* global_work_offset,
//                const size_t* global_work_size,
//                const size_t* local_work_size,
//                cl_uint num_events_in_wait_list,
//                const cl_event* event_wait_list,
//                cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
//    if ((error = dlerror()) != NULL) {
//        fputs(error, stderr);
//        exit(1);
//    }
//}
//
//void* ArmCLWorkGroupScheduler::ThreadFunc(void* argp)
//{
//  ((ArmCLWorkGroupScheduler*)argp)->Run();
//  return NULL;
//}

ArmCLCommandQueue::ArmCLCommandQueue(cl_command_queue command_queue, ArmCLContext* armcl_context)
    : command_queue_(command_queue)
{
    context_ = armcl_context->context();
    //cpu_context_ = armcl_context->cpu_context();

    armcl_context_ = armcl_context;

    PrepareOpenCLCalls();

    cl_int err;

    //gpu_work_group_pool_ = new ArmCLWorkGroupPool(num_gpu_compute_units_,
    //        CL_DEVICE_TYPE_GPU);
    //cpu_work_group_pool_ = new ArmCLWorkGroupPool(num_cpu_compute_units_,
    //        CL_DEVICE_TYPE_CPU);

    //gpu_workgroup_scheduler_ = new ArmCLWorkGroupScheduler(
    //    num_gpu_compute_units_,
    //    cpu_work_group_pool_,
    //    gpu_work_group_pool_,
    //    CL_DEVICE_TYPE_GPU);

    //cpu_workgroup_scheduler_ = new ArmCLWorkGroupScheduler(
    //    num_cpu_compute_units_,
    //    cpu_work_group_pool_,
    //    gpu_work_group_pool_,
    //    CL_DEVICE_TYPE_CPU);

//#ifdef AMD
//    profiler_ = new MCTProfiler(1);
//    profiler_->StartProfiling();
//#endif
}

ArmCLCommandQueue::~ArmCLCommandQueue()
{
}

cl_int ArmCLCommandQueue::CoRun(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    const char* kernel_name = armcl_kernel->kernel_name();
    std::cout << "CoRun " << kernel_name << std::endl;

    int num_cpu_cores = atoi(std::getenv("NUM_CPU_CORES"));
    int throttling_mod = atoi(std::getenv("THROTTLING_MOD"));
    int throttling_alloc = atoi(std::getenv("THROTTLING_ALLOC"));

    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));

    if (num_cpu_cores == 0 && throttling_alloc == 0) {
        return -1;
    } else if (num_cpu_cores == 0) {
        cpu_ratio = 0;
        gpu_ratio = 100;
    } else if (throttling_alloc == 0) {
        cpu_ratio = 100;
        gpu_ratio = 0;
    }

    size_t total_num_wgs;
    size_t gpu_partitioned_global_work_size[3] = { 0, };
    size_t gpu_wgs;

    size_t alloc_index;
    size_t cpu_wgs;
    size_t alloc_wgs;

    if (work_dim == 3) {
        size_t num_wgs[3];
        num_wgs[0] = global_work_size[0]/local_work_size[0];
        num_wgs[1] = global_work_size[1]/local_work_size[1];
        num_wgs[2] = global_work_size[2]/local_work_size[2];
        total_num_wgs = num_wgs[0] * num_wgs[1] * num_wgs[2];

        size_t gpu_wgs_0 = (size_t)((double)num_wgs[0] * ((double)gpu_ratio / 100.0));
        gpu_partitioned_global_work_size[0] = gpu_wgs_0 * local_work_size[0];
        gpu_partitioned_global_work_size[1] = global_work_size[1];
        gpu_partitioned_global_work_size[2] = global_work_size[2];
        gpu_wgs = gpu_wgs_0 * num_wgs[1] * num_wgs[2];

        size_t cpu_wgs_0 = num_wgs[0] - gpu_wgs_0;
        alloc_index = gpu_wgs_0 * num_wgs[1] * num_wgs[2];
        cpu_wgs = cpu_wgs_0 * num_wgs[1] * num_wgs[2];
        if (cpu_wgs) alloc_wgs = cpu_wgs / num_cpu_cores;
    } else if (work_dim == 2) {
        size_t num_wgs[2];
        num_wgs[0] = global_work_size[0]/local_work_size[0];
        num_wgs[1] = global_work_size[1]/local_work_size[1];
        total_num_wgs = num_wgs[0] * num_wgs[1];

        size_t gpu_wgs_0 = (size_t)((double)num_wgs[0] * ((double)gpu_ratio / 100.0));
        gpu_partitioned_global_work_size[0] = gpu_wgs_0 * local_work_size[0];
        gpu_partitioned_global_work_size[1] = global_work_size[1];
        gpu_wgs = gpu_wgs_0 * num_wgs[1];

        size_t cpu_wgs_0 = num_wgs[0] - gpu_wgs_0;
        alloc_index = gpu_wgs_0 * num_wgs[1];
        cpu_wgs = cpu_wgs_0 * num_wgs[1];
        if (cpu_wgs) alloc_wgs = cpu_wgs / num_cpu_cores;
    } else if (work_dim == 1) {
        total_num_wgs = global_work_size[0]/local_work_size[0];
        gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
        gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];

        alloc_index = gpu_wgs;
        cpu_wgs = total_num_wgs - gpu_wgs;
        if (cpu_wgs) alloc_wgs = cpu_wgs / num_cpu_cores;
    }

    if (gpu_wgs) {
        cl_kernel throttling_kernel = armcl_kernel->throttling_kernel();
        cl_int throttling_kernel_num_args = armcl_kernel->throttling_kernel_num_args();

        cl_int err = clSetKernelArg(throttling_kernel,
                throttling_kernel_num_args-2,
                sizeof(int),
                &throttling_mod);
        if (err != CL_SUCCESS) {
            std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;
        }

        err = clSetKernelArg(throttling_kernel,
                throttling_kernel_num_args-1,
                sizeof(int),
                (void*)&throttling_alloc);
        if (err != CL_SUCCESS) {
            std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                throttling_kernel,
                work_dim,
                global_work_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
        if (err != CL_SUCCESS) {
            std::cerr << "ArmCL: clEnqueueNDRangeKernel for throttling_kernel" << std::endl;
        }
    }

    if (cpu_wgs) {
        std::vector<std::thread*> threads;

        int (*cpu_thread)(int id, void* argp);
        cpu_thread = armcl_kernel->throttling_kernel_cpu_thread();
        int cpu_thread_num_args = armcl_kernel->throttling_kernel_cpu_thread_num_args();
        unsigned long* cpu_thread_params = armcl_kernel->throttling_kernel_cpu_thread_params();

        for (int i = 0; i < num_cpu_cores; i++) {
            int base = i*cpu_thread_num_args;

            if (i == num_cpu_cores - 1) {
                cpu_thread_params[base + cpu_thread_num_args-4] = (unsigned long)global_work_size;
                cpu_thread_params[base + cpu_thread_num_args-3] = (unsigned long)local_work_size;
                cpu_thread_params[base + cpu_thread_num_args-2] = (unsigned long)alloc_index;
                cpu_thread_params[base + cpu_thread_num_args-1] = (unsigned long)(total_num_wgs - alloc_index);

                std::thread* thread_ = new std::thread(cpu_thread, i, (void*)NULL);
                threads.push_back(thread_);
            } else {
                cpu_thread_params[base + cpu_thread_num_args-4] = (unsigned long)global_work_size;
                cpu_thread_params[base + cpu_thread_num_args-3] = (unsigned long)local_work_size;
                cpu_thread_params[base + cpu_thread_num_args-2] = (unsigned long)alloc_index;
                cpu_thread_params[base + cpu_thread_num_args-1] = (unsigned long)alloc_wgs;

                std::thread* thread_ = new std::thread(cpu_thread, i, (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++)
            threads[i]->join();
    }

    cl_int err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL MBAP: clFinish error" << std::endl;
    }

    return CL_SUCCESS;
}

inline int wg_first(std::atomic_int* worklist, int alloc_wgs)
{
    int current = worklist->fetch_add(alloc_wgs);
    return current;
}

inline int wg_next(std::atomic_int* worklist, int alloc_wgs)
{
    int current = worklist->fetch_add(alloc_wgs);
    return current;
}

inline bool wg_more(int current, int n_tasks)
{
    return current < n_tasks;
}

double scaleinfo[11][2] = {
	{1.4968051118210863,0.4999943307476847},
	{702215.769968051,26210097.87878559},
	{160.0,96.00087134500272},
	{0.8961661341853036,1.1376967282667845},
	{1.6421725239616614,0.7789703418942893},
	{0.4073482428115016,0.49134512751545045},
	{0.4057507987220447,0.49104120507538546},
	{27.926517571884983,20.302870631233063},
	{10.180511182108626,18.658121350906967},
	{0.5113636363636364,0.34933103769942764},
	{0.5113636363636364,0.3173719725077304},
};

struct info {
	float time;
	int core;
};

void predict(struct svm_model* model, svm_node* features, std::vector<struct info>& result, int fnum, int tid, const int numCPU)
{
	struct svm_node* _features = (struct svm_node *) malloc((fnum+1) * sizeof(struct svm_node));
	std::copy_n(features, fnum+1, _features);

	double bestTime = -100;
	int bestCore=0;
	_features[fnum].index = -1;
	int repeat = (numCPU+2) * 8 / numCPU;
	for(int cnum=tid*repeat; cnum<(tid+1)*repeat; cnum++) {
		int cpu = cnum/9;
		int gpu = cnum%9;
		if(cpu == 0 && gpu == 0)
			continue;
		if(cpu > 4)
			continue;
		_features[fnum-3] = {fnum-2, (cpu/4.0-scaleinfo[9][0])/scaleinfo[9][1]};
		_features[fnum-2] = {fnum-1, (gpu/8.0-scaleinfo[10][0])/scaleinfo[10][1]};
		double predict = svm_predict(model, _features);
		if(predict > bestTime) {
			bestTime = predict;
			bestCore = cnum;
		}
	}
	result[tid] = {bestTime, bestCore};
	free(_features);
}

void predict_tree(std::vector<double> features, std::vector<struct info>& result, int tid, const int numCPU)
{
	double bestTime = -100;
	int bestCore=0;
	int repeat = (numCPU+2) * 8 / numCPU;
	int fnum = features.size();
	for(int cnum=tid*repeat; cnum<(tid+1)*repeat; cnum++) {
		int cpu = cnum/9;
		int gpu = cnum%9;
		if(cpu == 0 && gpu == 0)
			continue;
		if(cpu > 4)
			continue;
		features[fnum-2] = cpu/(double)numCPU;
		features[fnum-1] = gpu/8.0;

		double predict = dopia_decision_tree_skylake_DT_UL(features);
		if(predict > bestTime) {
			bestTime = predict;
			bestCore = cnum;
		}
	}
	result[tid] = {bestTime, bestCore};
}

cl_int ArmCLCommandQueue::DynCoRun(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    const char* kernel_name = armcl_kernel->kernel_name();
    std::cout << "DynCoRun " << kernel_name << std::endl;
/*
    int num_cpu_cores = atoi(std::getenv("NUM_CPU_CORES"));
    int throttling_mod = atoi(std::getenv("THROTTLING_MOD"));
    int throttling_alloc = atoi(std::getenv("THROTTLING_ALLOC"));
	*/
	int num_cpu_cores, throttling_alloc;
	int throttling_mod = 8;

	// Get parameters from ML Model
	{
		struct timespec start, finish;
		std::string _kernel_name(kernel_name);
	/*
		std::string model_path(std::getenv("MODEL_PATH"));
		struct svm_model* model;
		model = svm_load_model(model_path.c_str());
		if(model == 0) {
			std::cout << "ERROR: model not exists" << std::endl;
			exit(0);
		}
		*/
		std::string armcl_root(std::getenv("ARMCL_ROOT"));
		std::string cmd = armcl_root + "/runtime/extract_features.py " + _kernel_name + " "
					 + std::to_string(work_dim) + " " + std::to_string(*global_work_size) + " "
					 + std::to_string(*local_work_size) + " " + "skylake";
		FILE* pipe = popen(cmd.c_str(), "r");
		char out[256];

		// if(!pipe) return ??
		int fnum = 11;
		std::vector<double> features(fnum);
		for(int i=0; i<fnum-2; i++) {
			double value;
			fscanf(pipe, "%lf", &value);
			features[i] = value;
		}
		clock_gettime(CLOCK_MONOTONIC, &start);
		const int numCPU = 4; //(int) std::thread::hardware_concurrency();
		std::vector<std::thread*> threads;
		std::vector<struct info> result (8);

		for(int tid = 0; tid < numCPU; tid++) {
			std::thread* thread_ = new std::thread(predict_tree, features, std::ref(result), tid, numCPU);
			threads.push_back(thread_);
		}

		for (int i = 0; i < numCPU; i++)
			threads[i]->join();

		int bestCPU=0, bestGPU=0;
		double bestTime = -100;
		for(int tid=0; tid < numCPU; tid++) {
			if(result[tid].time > bestTime) {
				bestCPU = result[tid].core/9;
				bestGPU = result[tid].core%9;
			}
		}

		/*
		struct svm_node* features = (struct svm_node *) malloc((fnum+1) * sizeof(struct svm_node));

		for(int i=0; i<fnum-2; i++) {
			double value;
			fscanf(pipe, "%lf", &value);
			value = (value-scaleinfo[i][0])/scaleinfo[i][1];
			features[i] = {i+1, value};
		}
		features[fnum].index = -1;

		clock_gettime(CLOCK_MONOTONIC, &start);

		std::vector<std::thread*> threads;
		std::vector<struct info> result (8);

		for(int tid = 0; tid < numCPU; tid++) {
			std::thread* thread_ = new std::thread(predict, model, features, std::ref(result), fnum, tid, numCPU);
			threads.push_back(thread_);
		}
		for (int i = 0; i < numCPU; i++)
			threads[i]->join();

		int bestCPU=0, bestGPU=0;
		double bestTime = -100;
		for(int tid=0; tid < numCPU; tid++) {
			if(result[tid].time > bestTime) {
				bestCPU = result[tid].core/9;
				bestGPU = result[tid].core%9;
			}
		}
		pclose(pipe);
		free(features);
		*/

		num_cpu_cores = bestCPU;
		throttling_alloc = bestGPU;
		clock_gettime(CLOCK_MONOTONIC, &finish);
	}

    if (num_cpu_cores == 0 && throttling_alloc == 0)
        return -1;
    else if (num_cpu_cores == 0)
        return GPURun(armcl_kernel, work_dim,
                global_work_offset, global_work_size, local_work_size,
                num_events_in_wait_list, event_wait_list, event);
    else if (throttling_alloc == 0)
        return CPURun(armcl_kernel, work_dim,
                global_work_offset, global_work_size, local_work_size,
                num_events_in_wait_list, event_wait_list, event);

    std::atomic_int* worklist = new std::atomic_int(0);
    size_t num_tasks = global_work_size[0]/local_work_size[0];

    // Launch CPU threads
    std::vector<std::thread*> threads;

    int (*cpu_thread)(int id, void* argp);
    cpu_thread = armcl_kernel->dynthrottling_kernel_cpu_thread();
    int cpu_thread_num_args = armcl_kernel->dynthrottling_kernel_cpu_thread_num_args();
    unsigned long* cpu_thread_params = armcl_kernel->dynthrottling_kernel_cpu_thread_params();

    for (int i = 0; i < num_cpu_cores; i++) {
        int base = i*cpu_thread_num_args;
        cpu_thread_params[base + cpu_thread_num_args-4] = (unsigned long)global_work_size;
        cpu_thread_params[base + cpu_thread_num_args-3] = (unsigned long)local_work_size;
        cpu_thread_params[base + cpu_thread_num_args-2] = (unsigned long)worklist;
        cpu_thread_params[base + cpu_thread_num_args-1] = (unsigned long)num_tasks;

        std::thread* thread_ = new std::thread(cpu_thread, i, (void*)NULL);
        threads.push_back(thread_);
    }

    cl_int err;
    cl_kernel throttling_kernel = armcl_kernel->throttling_kernel();
    cl_int throttling_kernel_num_args = armcl_kernel->throttling_kernel_num_args();
    err = clSetKernelArg(throttling_kernel, throttling_kernel_num_args-2, sizeof(int), (void*)&throttling_mod);
    if (err != CL_SUCCESS) std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;
    err = clSetKernelArg(throttling_kernel, throttling_kernel_num_args-1, sizeof(int), (void*)&throttling_alloc);
    if (err != CL_SUCCESS) std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;

    // Launch GPU kernels
    size_t alloc_gpu = (size_t)(num_tasks / 10);
    for (int wg_id = wg_first(worklist, alloc_gpu); wg_more(wg_id, num_tasks); wg_id = wg_next(worklist, alloc_gpu)) {
        if (wg_id + alloc_gpu >= num_tasks ) alloc_gpu = num_tasks - wg_id;
        if (alloc_gpu <= 0) continue;
        size_t gpu_global_work_offset[3] = { 0, };
        size_t gpu_global_work_size[3] = { 0, };
        if (work_dim == 3) {
            gpu_global_work_size[0] = alloc_gpu * local_work_size[0];
            gpu_global_work_size[1] = global_work_size[1];
            gpu_global_work_size[2] = global_work_size[2];
            gpu_global_work_offset[0] = wg_id * local_work_size[0];
            gpu_global_work_offset[1] = 0;
            gpu_global_work_offset[2] = 0;
        } else if (work_dim == 2) {
            gpu_global_work_size[0] = alloc_gpu * local_work_size[0];
            gpu_global_work_size[1] = global_work_size[1];
            gpu_global_work_offset[0] = wg_id * local_work_size[0];
            gpu_global_work_offset[1] = 0;
        } else if (work_dim == 1) {
            gpu_global_work_size[0] = alloc_gpu * local_work_size[0];
            gpu_global_work_offset[0] = wg_id * local_work_size[0];
        }

        cl_event gpu_event;
        cl_int err = clEnqueueNDRangeKernel_p_(command_queue_,
                throttling_kernel,
                work_dim,
                gpu_global_work_offset,
                gpu_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                &gpu_event);
        if (err != CL_SUCCESS) {
            std::cerr << "ArmCL: clEnqueueNDRangeKernel_p_ for throttling_kernel" << std::endl;
        }

        clWaitForEvents(1, &gpu_event);
    }

    for (int i = 0; i < num_cpu_cores; i++)
        threads[i]->join();

    return CL_SUCCESS;
}

cl_int ArmCLCommandQueue::CPURun(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    const char* kernel_name = armcl_kernel->kernel_name();
    std::cout << "CPURun " << kernel_name << std::endl;

    int num_cpu_cores = atoi(std::getenv("NUM_CPU_CORES"));
    int throttling_mod = atoi(std::getenv("THROTTLING_MOD"));
    int throttling_alloc = atoi(std::getenv("THROTTLING_ALLOC"));

    size_t total_num_wgs;
    size_t alloc_index = 0; // assume global offset is 0
    size_t alloc_wgs;

    if (work_dim == 3) {
        size_t num_wgs[3];
        num_wgs[0] = global_work_size[0]/local_work_size[0];
        num_wgs[1] = global_work_size[1]/local_work_size[1];
        num_wgs[2] = global_work_size[2]/local_work_size[2];
        total_num_wgs = num_wgs[0] * num_wgs[1] * num_wgs[2];
        alloc_wgs = total_num_wgs / num_cpu_cores;
    } else if (work_dim == 2) {
        size_t num_wgs[2];
        num_wgs[0] = global_work_size[0]/local_work_size[0];
        num_wgs[1] = global_work_size[1]/local_work_size[1];
        total_num_wgs = num_wgs[0] * num_wgs[1];
        alloc_wgs = total_num_wgs / num_cpu_cores;
    } else if (work_dim == 1) {
        total_num_wgs = global_work_size[0]/local_work_size[0];
        alloc_wgs = total_num_wgs / num_cpu_cores;
    }

    std::vector<std::thread*> threads;

    int (*cpu_thread)(int id, void* argp);
    cpu_thread = armcl_kernel->throttling_kernel_cpu_thread();
    int cpu_thread_num_args = armcl_kernel->throttling_kernel_cpu_thread_num_args();
    unsigned long* cpu_thread_params = armcl_kernel->throttling_kernel_cpu_thread_params();

    for (int i = 0; i < num_cpu_cores; i++) {
        int base = i*cpu_thread_num_args;

        if (i == num_cpu_cores - 1) {
            cpu_thread_params[base + cpu_thread_num_args-4] = (unsigned long)global_work_size;
            cpu_thread_params[base + cpu_thread_num_args-3] = (unsigned long)local_work_size;
            cpu_thread_params[base + cpu_thread_num_args-2] = (unsigned long)alloc_index;
            cpu_thread_params[base + cpu_thread_num_args-1] = (unsigned long)(total_num_wgs - alloc_index);

            std::thread* thread_ = new std::thread(cpu_thread, i, (void*)NULL);
            threads.push_back(thread_);
        } else {
            cpu_thread_params[base + cpu_thread_num_args-4] = (unsigned long)global_work_size;
            cpu_thread_params[base + cpu_thread_num_args-3] = (unsigned long)local_work_size;
            cpu_thread_params[base + cpu_thread_num_args-2] = (unsigned long)alloc_index;
            cpu_thread_params[base + cpu_thread_num_args-1] = (unsigned long)alloc_wgs;

            std::thread* thread_ = new std::thread(cpu_thread, i, (void*)NULL);
            threads.push_back(thread_);
        }

        alloc_index += alloc_wgs;
    }

    for (int i = 0; i < num_cpu_cores; i++)
        threads[i]->join();

    return CL_SUCCESS;
}

cl_int ArmCLCommandQueue::GPURun(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    const char* kernel_name = armcl_kernel->kernel_name();
    std::cout << "GPURun " << kernel_name << std::endl;

    int throttling_mod = atoi(std::getenv("THROTTLING_MOD"));
    int throttling_alloc = atoi(std::getenv("THROTTLING_ALLOC"));

    cl_kernel throttling_kernel = armcl_kernel->throttling_kernel();
    cl_int throttling_kernel_num_args = armcl_kernel->throttling_kernel_num_args();

    cl_int err = clSetKernelArg(throttling_kernel,
            throttling_kernel_num_args-2,
            sizeof(int),
            &throttling_mod);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;
    }

    err = clSetKernelArg(throttling_kernel,
            throttling_kernel_num_args-1,
            sizeof(int),
            (void*)&throttling_alloc);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL: clSetKernelArg for throttling_kernel" << std::endl;
    }

    err = clEnqueueNDRangeKernel_p_(command_queue_,
            throttling_kernel,
            work_dim,
            global_work_offset,
            global_work_size,
            local_work_size,
            num_events_in_wait_list,
            event_wait_list,
            event);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL: clEnqueueNDRangeKernel for throttling_kernel" << std::endl;
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL MBAP: clFinish error" << std::endl;
    }

    return CL_SUCCESS;
}

cl_int ArmCLCommandQueue::OnTheFly(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "empty" << std::endl;
}

cl_int ArmCLCommandQueue::Optimal(
        ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "empty" << std::endl;
}

void ArmCLCommandQueue::PrepareOpenCLCalls()
{
    char* error;

    clCreateCommandQueue_p_ = (cl_command_queue CL_API_CALL (*) (
                cl_context context,
                cl_device_id device,
                cl_command_queue_properties properties,
                cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateCommandQueue");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueNDRangeKernel_p_ = (cl_int (*) (cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clSetKernelArg_p_ = (cl_int (*) (cl_kernel kernel,
                cl_uint arg_index,
                size_t arg_size,
                const void* arg_value))
        dlsym(RTLD_NEXT, "clSetKernelArg");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clCreateBuffer_p_ = (cl_mem (*) (cl_context context,
                cl_mem_flags flags,
                size_t size,
                void* host_ptr,
                cl_int* errcode_ret))
        dlsym(RTLD_NEXT, "clCreateBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueWriteBuffer_p_ = (cl_int CL_API_CALL (*) (
                cl_command_queue command_queue,
                cl_mem buffer,
                cl_bool blocking_write,
                size_t offset,
                size_t size,
                const void* ptr,
                cl_uint num_events_in_wait_list,
                const cl_event* event_wait_list,
                cl_event* event))dlsym(RTLD_NEXT, "clEnqueueWriteBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clGetEventInfo_p_ = (cl_int CL_API_CALL (*) (
                cl_event event,
                cl_profiling_info param_name,
                size_t param_value_size,
                void* param_value,
                size_t* param_value_size_ret))dlsym(RTLD_NEXT, "clGetEventInfo");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }
}
