//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#ifndef __ARM_OpenCL_Command_Queue_HEADER_
#define __ARM_OpenCL_Command_Queue_HEADER_

#include <CL/cl.h>
#include "ArmCLContext.hpp"
#include "ArmCLKernel.hpp"

#include <deque>
//#include <pthread.h>
#include <semaphore.h>
#include <thread>

//typedef struct _wgrange {
//    size_t start;
//    size_t end;
//    size_t nwg;
//} wgrange;
//
//typedef struct _ArmCLWorkGroupAssignment {
//    cl_command_queue command_queue;
//    cl_kernel kernel;
//    cl_uint work_dim;
//    size_t local_work_size;
//    cl_uint num_events_in_wait_list;
//    const cl_event* event_wait_list;
//    cl_event* event;
//} ArmCLWorkGroupAssignment;
//
//class ArmCLWorkGroupPool {
//  public:
//    ArmCLWorkGroupPool(int num_cus, cl_device_type device_type);
//    ~ArmCLWorkGroupPool();
//
//    void lock(void) { pthread_mutex_lock(&mutex_); }
//    void unlock(void) { pthread_mutex_unlock(&mutex_); }
//
//    void set_alloc_unit(size_t chunk_size) { chunk_size_ = chunk_size; }
//
//    wgrange* next_work(void);
//    wgrange* next_work_chunk(void);
//    wgrange* pop_work(void);
//    wgrange* pop_work(size_t num_wgs);
//    wgrange* pop_front(void);
//    wgrange* pop_back(void);
//
//    bool work_stealing(ArmCLWorkGroupPool* other_wg_pool);
//    void push_work(wgrange* wg_range);
//
//  private:
//    void coalesce_ranges(void);
//
//    const int num_cus_;
//    const cl_device_type device_type_;
//
//    size_t chunk_size_;
//
//    std::map<int, wgrange*> wg_ranges_;
//    int num_ranges_;
//    int wg_total_;
//    bool is_done_;
//    pthread_mutex_t mutex_;
//};
//
//class ArmCLWorkGroupScheduler {
//  public:
//    ArmCLWorkGroupScheduler(size_t num_available_cus,
//            ArmCLWorkGroupPool* cpu_wg_pool,
//            ArmCLWorkGroupPool* gpu_wg_pool,
//            cl_device_type device_type);
//    ~ArmCLWorkGroupScheduler();
//
//    bool IsIdle() const;
//    void Join();
//    void Launch(ArmCLWorkGroupAssignment* wga);
//
//  private:
//    void PrepareOpenCLCalls();
//    void Run();
//
//    cl_device_type device_type_;
//
//    pthread_t thread_;
//    bool thread_running_;
//
//    const int num_available_cus_;
//    ArmCLWorkGroupPool* cpu_wg_pool_;
//    ArmCLWorkGroupPool* gpu_wg_pool_;
//
//    ArmCLWorkGroupAssignment* wga_;
//
//    bool schedule_to_do_;
//
//    pthread_mutex_t mutex_schedule_queue_;
//    pthread_cond_t cond_schedule_queue_;
//
//    cl_int CL_API_CALL (*clEnqueueNDRangeKernel_p_) (
//            cl_command_queue command_queue,
//            cl_kernel kernel,
//            cl_uint work_dim,
//            const size_t* global_work_offset,
//            const size_t* global_work_size,
//            const size_t* local_work_size,
//            cl_uint num_events_in_wait_list,
//            const cl_event* event_wait_list,
//            cl_event* event);
//
//  public:
//    static void* ThreadFunc(void* argp);
//};

class ArmCLCommandQueue {
  public:
    ArmCLCommandQueue(cl_command_queue command_queue,
            ArmCLContext* armcl_context);
    ~ArmCLCommandQueue();

    cl_command_queue command_queue() const { return command_queue_; }

    cl_int CoRun(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int DynCoRun(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int CPURun(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int GPURun(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int OnTheFly(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int Optimal(ArmCLKernel* armcl_kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

  private:
    void PrepareOpenCLCalls();

    cl_context context_;
    cl_context cpu_context_;

    cl_command_queue command_queue_;

    //ArmCLWorkGroupPool* cpu_work_group_pool_;
    //ArmCLWorkGroupPool* gpu_work_group_pool_;
    //ArmCLWorkGroupScheduler* gpu_workgroup_scheduler_;
    //ArmCLWorkGroupScheduler* cpu_workgroup_scheduler_;

    ArmCLContext* armcl_context_;

    cl_command_queue CL_API_CALL (*clCreateCommandQueue_p_) (
            cl_context context,
            cl_device_id device,
            cl_command_queue_properties properties,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueNDRangeKernel_p_) (
            cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int CL_API_CALL (*clSetKernelArg_p_)(cl_kernel kernel,
            cl_uint arg_index,
            size_t arg_size,
            const void* arg_value);

    cl_mem CL_API_CALL (*clCreateBuffer_p_)(cl_context context,
            cl_mem_flags flags,
            size_t size,
            void* host_ptr,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueWriteBuffer_p_)(
            cl_command_queue command_queue,
            cl_mem buffer,
            cl_bool blocking_write,
            size_t offset,
            size_t size,
            const void* ptr,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int CL_API_CALL (*clGetEventInfo_p_)(
            cl_event event,
            cl_profiling_info param_name,
            size_t param_value_size,
            void* param_value,
            size_t* param_value_size_ret);
};

#endif

