//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLContext.hpp"
#include <thread>

#define CHECK_ERROR(err)                                   \
	if (err != CL_SUCCESS) {                                 \
		printf("ArmCL: [%s:%d] error %d\n", __FILE__, __LINE__, err); \
		exit(EXIT_FAILURE);                                    \
	}

void ArmCLProgramSource::AddSource(const char* source)
{
    sources_.push_back(std::string(source));
    concat_source_.append(source);
}

void ArmCLProgramSource::AddSource(const char* source, size_t length)
{
    sources_.push_back(std::string(source, length));
    concat_source_.append(source, length);
}

void ArmCLProgramSource::SetHeaderName(const char* header_name)
{
    header_name_ = std::string(header_name);
}

static size_t PipeRead(const char* filename, char* buf, size_t size)
{
    FILE* fp = popen(filename, "r");
	std::cout << "why1" << std::endl;
    size_t read_size = fread(buf, sizeof(char), size, fp);
	std::cout << "why2" << std::endl;
    pclose(fp);
    return read_size;
}

static int GetTotalSystemCores()
{
    char buf[8];
    PipeRead("/bin/cat /sys/devices/system/cpu/present", buf, sizeof(buf) - 1);
    int from, to;
    sscanf(buf, "%d-%d", &from, &to);
    int num_cores = to - from + 1;
    std::cout << num_cores << " cores are detected" << std::endl;

    return num_cores;
}

ArmCLContext::ArmCLContext(cl_context context)
    : context_(context)
{
#ifdef DEBUG
    std::cout << "ArmCL: "
        << " context created"
        << std::endl;
#endif

    PrepareOpenCLCalls();
    PrepareCPUContext();
    QueryDeviceInfo();

    num_total_cpu_cores_ = (int) std::thread::hardware_concurrency();
	//num_total_cpu_cores_ = GetTotalSystemCores();
}

ArmCLContext::~ArmCLContext()
{
#ifdef DEBUG
    std::cout << "ArmCL: "
        << " context released"
        << std::endl;
#endif
}

void ArmCLContext::RegisterPrograms(ArmCLProgram* armcl_program)
{
    armcl_programs_.push_back(armcl_program);
}

void ArmCLContext::PreprocessingWithSource(cl_uint count,
        const char ** strings,
        const size_t * lengths,
        char ** generated_strings,
        size_t * generated_lengths)
{
#ifdef DEBUG
    std::cout << "ArmCL: "
        << " preprocessing with source"
        << std::endl;
#endif

    // concatenate input strings
    source_ = new ArmCLProgramSource();
    for (cl_uint i = 0; i < count; i++) {
        if (lengths == NULL || lengths[i] == 0)
            source_->AddSource(strings[i]);
        else
            source_->AddSource(strings[i], lengths[i]);
    }

    FILE* fp = fopen("temp.cl", "w");
    fprintf(fp, "%s", source_->concat_source());
    fclose(fp);

    // call preprocessor
    int ret = system("armcl-prep temp.cl");

    // forward preprocessed strings
}

void ArmCLContext::PrepareOpenCLCalls()
{
    char* error;

    clCreateContext_p_ = (cl_context (*) (
            const cl_context_properties* properties,
            cl_uint num_devices,
            const cl_device_id* devices,
            void (*pfn_notify) (
                const char* effinfo,
                const void* private_info,
                size_t cb,
                void* user_data
                ),
            void* user_data,
            cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateContext");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }
}

void ArmCLContext::PrepareCPUContext()
{
    cl_device_type DEV_TYPE = CL_DEVICE_TYPE_CPU;
    cl_platform_id* platforms;
    cl_uint num_platforms;
    size_t platform_name_size;
    char* platform_name;
    cl_platform_id armcl_platform = NULL;
    //cl_context_properties context_properties[3];
    cl_int err;

    err = clGetPlatformIDs(0, NULL, &num_platforms);
    CHECK_ERROR(err);

#ifdef DEBUG
    std::cout << "ArmCL: " << num_platforms << " platforms are detected."
        << std::endl;
#endif

    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id) * num_platforms);
    err = clGetPlatformIDs(num_platforms, platforms, NULL);
    CHECK_ERROR(err);

    for (int i = 0; i < num_platforms; i++) {
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 0, NULL,
                &platform_name_size);
        CHECK_ERROR(err);

        platform_name = (char*)malloc(sizeof(char) * platform_name_size);
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME,
                platform_name_size, platform_name, NULL);
        CHECK_ERROR(err);

#ifdef DEBUG
        std::cout << "ArmCL: " << "Platform " << i << " " << platform_name
            << std::endl;
#endif
        armcl_platform = platforms[i];
        free(platform_name);
    }

    if (armcl_platform == NULL) {
        std::cout << "ArmCL: 0 Platforms are detected." << std::endl;
        exit(1);
    }

//    err = clGetDeviceIDs(armcl_platform, DEV_TYPE, 1, &cpu_device_, NULL);
//    CHECK_ERROR(err);
//
//    cl_ulong global_mem_size;
//    err = clGetDeviceInfo(cpu_device_, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong),
//            &global_mem_size, 0);
//#ifdef DEBUG
//    std::cout << "ArmCL: global_mem_size: " << global_mem_size << std::endl;
//#endif
//
//    context_properties[0] = CL_CONTEXT_PLATFORM;
//    context_properties[1] = (cl_context_properties)armcl_platform;
//    context_properties[2] = 0;
//
//    cpu_context_ = clCreateContext_p_(context_properties, 1, &cpu_device_,
//            NULL, NULL, &err);
//    CHECK_ERROR(err);
}

void ArmCLContext::QueryDeviceInfo()
{
    cl_int err;

    err = clGetContextInfo(context_,
            CL_CONTEXT_DEVICES,
            sizeof(cl_device_id),
            &gpu_device_,
            NULL);

    err = clGetDeviceInfo(gpu_device_,
            CL_DEVICE_MAX_COMPUTE_UNITS,
            sizeof(cl_uint),
            &num_gpu_compute_units_,
            NULL);

//    err = clGetDeviceInfo(cpu_device_,
//            CL_DEVICE_MAX_COMPUTE_UNITS,
//            sizeof(cl_uint),
//            &num_cpu_compute_units_,
//            NULL);
//
//#ifdef DEBUG
//    std::cout << "ArmCL: num_gpu_compute_units: " << num_gpu_compute_units_
//        << " num_cpu_compute_units: " << num_cpu_compute_units_
//        << std::endl;
//#endif
}
