//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#ifndef __ARM_OpenCL_Kernel_HEADER_
#define __ARM_OpenCL_Kernel_HEADER_

#include <CL/cl.h>
#include <iostream>
#include <dlfcn.h>
#include <stdio.h>
#include <map>
#include <vector>
#include <list>

#include "ArmCLContext.hpp"
#include "ArmCLProgram.hpp"
#include "ArmCLBuffer.hpp"

class ArmCLKernel {
  public:
    ArmCLKernel(cl_kernel kernel, const char* kernel_name, ArmCLProgram* armcl_program);
    ~ArmCLKernel();

    const char* kernel_name() { return kernel_name_; }

    // kernel
    cl_kernel kernel() const { return kernel_; }
    cl_uint kernel_num_args() const { return kernel_num_args_; }
    int (*kernel_cpu_thread())(int, void*) const { return kernel_cpu_thread_; }
    cl_uint kernel_cpu_thread_num_args() const { return kernel_cpu_thread_num_args_; }
    unsigned long* kernel_cpu_thread_params() const { return kernel_cpu_thread_params_; }

    // throttling kernel
    cl_kernel throttling_kernel() const { return throttling_kernel_; }
    cl_uint throttling_kernel_num_args() const { return throttling_kernel_num_args_; }
    int (*throttling_kernel_cpu_thread())(int, void*) const { return throttling_kernel_cpu_thread_; }
    cl_uint throttling_kernel_cpu_thread_num_args() const { return throttling_kernel_cpu_thread_num_args_; }
    unsigned long* throttling_kernel_cpu_thread_params() const { return throttling_kernel_cpu_thread_params_; }
    int (*dynthrottling_kernel_cpu_thread())(int, void*) const { return dynthrottling_kernel_cpu_thread_; }
    cl_uint dynthrottling_kernel_cpu_thread_num_args() const { return dynthrottling_kernel_cpu_thread_num_args_; }
    unsigned long* dynthrottling_kernel_cpu_thread_params() const { return dynthrottling_kernel_cpu_thread_params_; }



    // onthefly partitioning
    cl_kernel onthefly_kernel() const { return onthefly_kernel_; }
    cl_uint onthefly_kernel_num_args() const { return onthefly_kernel_num_args_; }
    int (*onthefly_kernel_cpu_thread())(int, void*) const { return onthefly_kernel_cpu_thread_; }
    cl_uint onthefly_kernel_cpu_thread_num_args() const { return onthefly_kernel_cpu_thread_num_args_; }
    unsigned long* onthefly_kernel_cpu_thread_params() const { return onthefly_kernel_cpu_thread_params_; }

    cl_int SetKernelArgSVMPointer(cl_uint arg_index, const void* arg_value);
    cl_int SetKernelArg(cl_uint arg_index, size_t arg_size, const void* arg_value, std::map<cl_mem, ArmCLBuffer*> armcl_buffers);

  private:
    void PrepareOpenCLCalls();
    void PrepareContextInformation();

    void PrepareThrottlingKernels();
    void PrepareOnTheFlyKernels();

    ArmCLProgram* armcl_program_;
    ArmCLContext* armcl_context_;

    cl_program program_;
    cl_context context_;

    const char* kernel_name_;
    const cl_kernel kernel_;
    cl_uint kernel_num_args_;
    int kernel_cpu_thread_num_args_;
    int (*kernel_cpu_thread_)(int id, void* argp);
    unsigned long* kernel_cpu_thread_params_;

    int throttling_ = 1;
    cl_uint throttling_kernel_num_args_;
    cl_kernel throttling_kernel_;
    int throttling_kernel_cpu_thread_num_args_;
    int (*throttling_kernel_cpu_thread_)(int id, void* argp);
    unsigned long* throttling_kernel_cpu_thread_params_;
    int dynthrottling_kernel_cpu_thread_num_args_;
    int (*dynthrottling_kernel_cpu_thread_)(int id, void* argp);
    unsigned long* dynthrottling_kernel_cpu_thread_params_;

    int onthefly_ = 0;
    cl_uint onthefly_kernel_num_args_;
    cl_kernel onthefly_kernel_;
    int onthefly_kernel_cpu_thread_num_args_;
    int (*onthefly_kernel_cpu_thread_)(int id, void* argp);
    unsigned long* onthefly_kernel_cpu_thread_params_;
    int (*onthefly_kernel_cpu_thread_array_[100])(int id, void* argp);
    unsigned long** onthefly_kernel_cpu_thread_array_params_;
    int (*onthefly_kernel_cpu_profiler_)(int id, void* argp);
    int onthefly_kernel_cpu_profiler_num_args_;
    unsigned long* onthefly_kernel_cpu_profiler_params_;

    cl_kernel CL_API_CALL (*clCreateKernel_p_) (cl_program program,
            const char* kernel_name,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueNDRangeKernel_p_) (
            cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int CL_API_CALL (*clSetKernelArg_p_)(cl_kernel kernel,
            cl_uint arg_index,
            size_t arg_size,
            const void* arg_value);

    cl_mem CL_API_CALL (*clCreateBuffer_p_)(cl_context context,
            cl_mem_flags flags,
            size_t size,
            void* host_ptr,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueWriteBuffer_p_)(
            cl_command_queue command_queue,
            cl_mem buffer,
            cl_bool blocking_write,
            size_t offset,
            size_t size,
            const void* ptr,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);
};

#endif
