/// Profiler
/// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#include "event.h"

#include <stdlib.h>
#include <unistd.h>
#include <linux/perf_event.h>
#include <sys/epoll.h>

#include <list>
#include <map>
#include <thread>
#include <cstdio>
#include <fstream>
#include <iostream>

class PMUReqManager {
  public:
    PMUReqManager();
    ~PMUReqManager();

    void InitPerfEventAttr();
    void InitPerfEventAttr(
        std::pair<__u32, __u64>* attr_mapping,
        event_type* attr_type,
        int num_events);

    void GetCurrentCount(sc_event_values& val);
    void GetCurrentCount(sc_event_values& val,
        unsigned long* values);

    void OpenEventFDs();

    void OpenEventFDs(int subset_index);
    void OpenEventFDsByPID(pid_t pid);

    void CloseEventFDs();
    void CloseEventFDs(int subset_index);

    void RecordAllCounters();

  private:
    void ReadCounters(sc_event_values *values);

    int _fd[MAX_EVENTS][NUM_CORES]; ///< event-perf file descriptors
    struct perf_event_attr _attr[MAX_EVENTS]; ///< event-perf attributes
    event_type _event_type[MAX_EVENTS]; ///< event type information
    sc_event_values _cur; ///< recently stored count values
    //std::mutex _cur_mutex;
};

class AMDProfiler {
  public:
    AMDProfiler() {}
    ~AMDProfiler() {}

    virtual void ProfilerInfo(std::string path) = 0;
    virtual void StartProfiling(void) = 0;
    virtual void StopProfiling(char* prefix_c, char* path_c) = 0;
};

class MCTProfiler: public AMDProfiler {
  public:
    MCTProfiler(int num_memories);
    ~MCTProfiler();

    virtual void ProfilerInfo(std::string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(char* prefix_c, char* path_c);

  private:
    PMUReqManager*  _pmu_manager;  ///< pmu request manager

    std::pair<__u32, __u64>* _attr_mapping;
    event_type* _attr_type;
    sc_event_values _start;        ///< pmu counter values at StartProfiling()

    struct timespec start;

    int num_memories_;
};

class DCTProfiler: public AMDProfiler {
  public:
    DCTProfiler(int dct_num, int num_memories);
    ~DCTProfiler();

    virtual void ProfilerInfo(std::string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(char* prefix_c, char* path_c);

  private:
    PMUReqManager*  _pmu_manager;  ///< pmu request manager

    std::pair<__u32, __u64>* _attr_mapping;
    event_type* _attr_type;
    sc_event_values _start;        ///< pmu counter values at StartProfiling()

    int dct_num_;

    int num_memories_;
};

