/// Event description for AMD64 machine (AMD Opteron 6380)
/// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>
#ifndef __EVENT_H_
#define __EVENT_H_

#include <linux/perf_event.h>

#define TOTAL_CYCLES_TYPE            PERF_TYPE_HARDWARE
#define TOTAL_CYCLES_CONFIG          PERF_COUNT_HW_CPU_CYCLES
#define TOTAL_INSTRUCTIONS_TYPE      PERF_TYPE_HARDWARE
#define TOTAL_INSTRUCTIONS_CONFIG    PERF_COUNT_HW_INSTRUCTIONS
#define STALL_CYCLES_TYPE            PERF_TYPE_RAW
#define STALL_CYCLES_CONFIG          0xD1
#define L2_CACHE_MISSES_TYPE         PERF_TYPE_RAW
#define L2_CACHE_MISSES_CONFIG       0x53177E
#define LLC_ACCESSES_TYPE            6
#define LLC_ACCESSES_CONFIG          0x40053F7E0
#define LLC_MISSES_TYPE              6
#define LLC_MISSES_CONFIG            0x40053F7E1

#define DRAM_ACCESSES_TYPE                         6
#define DRAM_ACCESSES_TYPE_CONFIG                  0x533fe0
#define DRAM_ACCESSES_DCT0_PAGE_HIT_TYPE           6
#define DRAM_ACCESSES_DCT0_PAGE_HIT_CONFIG         0x5301e0
#define DRAM_ACCESSES_DCT0_PAGE_MISS_TYPE          6
#define DRAM_ACCESSES_DCT0_PAGE_MISS_CONFIG        0x5302e0
#define DRAM_ACCESSES_DCT0_PAGE_CONFLICT_TYPE      6
#define DRAM_ACCESSES_DCT0_PAGE_CONFLICT_CONFIG    0x5304e0
#define DRAM_ACCESSES_DCT1_PAGE_HIT_TYPE           6
#define DRAM_ACCESSES_DCT1_PAGE_HIT_CONFIG         0x5308e0
#define DRAM_ACCESSES_DCT1_PAGE_MISS_TYPE          6
#define DRAM_ACCESSES_DCT1_PAGE_MISS_CONFIG        0x5310e0
#define DRAM_ACCESSES_DCT1_PAGE_CONFLICT_TYPE      6
#define DRAM_ACCESSES_DCT1_PAGE_CONFLICT_CONFIG    0x5320e0

#define MEMORY_CONTROLLER_REQUESTS_TYPE                               6
#define MEMORY_CONTROLLER_REQUESTS_CONFIG                             0x10053fff0
#define MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_TYPE         6
#define MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_CONFIG       0x1005301f0
#define MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_TYPE          6
#define MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_CONFIG        0x1005302f0
#define MEMORY_CONTROLLER_REQUESTS_PREFETCH_REQUESTS_TO_DCT_TYPE      6
#define MEMORY_CONTROLLER_REQUESTS_PREFETCH_REQUESTS_TO_DCT_CONFIG    0x1005304f0
#define MEMORY_CONTROLLER_REQUESTS_32_BYTES_SIZED_WRITES_TYPE         6
#define MEMORY_CONTROLLER_REQUESTS_32_BYTES_SIZED_WRITES_CONFIG       0x1005308f0
#define MEMORY_CONTROLLER_REQUESTS_64_BYTES_SIZED_WRITES_TYPE         6
#define MEMORY_CONTROLLER_REQUESTS_64_BYTES_SIZED_WRITES_CONFIG       0x1005310f0
#define MEMORY_CONTROLLER_REQUESTS_32_BYTES_SIZED_READS_TYPE          6
#define MEMORY_CONTROLLER_REQUESTS_32_BYTES_SIZED_READS_CONFIG        0x1005320f0
#define MEMORY_CONTROLLER_REQUESTS_64_BYTE_SIZED_READS_TYPE           6
#define MEMORY_CONTROLLER_REQUESTS_64_BYTE_SIZED_READS_CONFIG         0x1005340f0
#define MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_WHILE_WRITES_PENDING_TYPE      6
#define MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_WHILE_WRITES_PENDING_CONFIG    0x1005380f0

#define L3_CACHE_MISSES_TYPE             6
#define L3_CACHE_MISSES_CONFIG           0x40053F7E1
#define L3_CACHE_MISSES_CORE_0_TYPE      6
#define L3_CACHE_MISSES_CORE_0_CONFIG    0x4005307e1
#define L3_CACHE_MISSES_CORE_1_TYPE      6
#define L3_CACHE_MISSES_CORE_1_CONFIG    0x4005317e1
#define L3_CACHE_MISSES_CORE_2_TYPE      6
#define L3_CACHE_MISSES_CORE_2_CONFIG    0x4005327e1
#define L3_CACHE_MISSES_CORE_3_TYPE      6
#define L3_CACHE_MISSES_CORE_3_CONFIG    0x4005337e1
#define L3_CACHE_MISSES_CORE_4_TYPE      6
#define L3_CACHE_MISSES_CORE_4_CONFIG    0x4005347e1
#define L3_CACHE_MISSES_CORE_5_TYPE      6
#define L3_CACHE_MISSES_CORE_5_CONFIG    0x4005357e1
#define L3_CACHE_MISSES_CORE_6_TYPE      6
#define L3_CACHE_MISSES_CORE_6_CONFIG    0x4005367e1
#define L3_CACHE_MISSES_CORE_7_TYPE      6
#define L3_CACHE_MISSES_CORE_7_CONFIG    0x4005377e1

#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_ALL_TYPE    6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_ALL_CONFIG  0x10053ffe0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_0_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_0_CONFIG    0x1005301E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_1_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_1_CONFIG    0x1005302E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_2_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_2_CONFIG    0x1005304E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_3_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_3_CONFIG    0x1005308E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_4_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_4_CONFIG    0x1005310E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_5_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_5_CONFIG    0x1005320E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_6_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_6_CONFIG    0x1005340E0
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_7_TYPE      6
#define CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_7_CONFIG    0x1005380E0

#define NUM_CORES                    4
#define NUM_CORES_IN_NODE            4
#define MAX_EVENTS                   5

/// @brief event type
enum event_type {
  PER_CORE,    // per core event
  PER_CLUSTER, // per cluster event
};

/// @brief performance counter values
///
/// @description store performance counter in this format
typedef struct {
  unsigned long values[MAX_EVENTS][NUM_CORES];
  unsigned long timestamp[MAX_EVENTS][NUM_CORES];
} sc_event_values;

#endif // __EVENT_H_

