/// Profiler
/// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#include "event.h"
#include "pmu_interface.h"

#include <map>
#include <thread>
#include <cstdio>
#include <fstream>
#include <iostream>

class AMDProfiler {
  public:
    AMDProfiler() {}
    ~AMDProfiler() {}

    virtual void ProfilerInfo(std::string path) = 0;
    virtual void StartProfiling(void) = 0;
    virtual void StopProfiling(std::string prefix, std::string path) = 0;
};

class MCTProfiler: public AMDProfiler {
  public:
    MCTProfiler(int num_memories);
    ~MCTProfiler();

    virtual void ProfilerInfo(std::string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(std::string prefix, std::string path);

  private:
    PMUReqManager*  _pmu_manager;  ///< pmu request manager

    std::pair<__u32, __u64>* _attr_mapping;
    event_type* _attr_type;
    sc_event_values _start;        ///< pmu counter values at StartProfiling()

    struct timespec start;

    int num_memories_;
};

class DCTProfiler: public AMDProfiler {
  public:
    DCTProfiler(int dct_num, int num_memories);
    ~DCTProfiler();

    virtual void ProfilerInfo(std::string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(std::string prefix, std::string path);

  private:
    PMUReqManager*  _pmu_manager;  ///< pmu request manager

    std::pair<__u32, __u64>* _attr_mapping;
    event_type* _attr_type;
    sc_event_values _start;        ///< pmu counter values at StartProfiling()

    int dct_num_;

    int num_memories_;
};

