// 1Mat3DAdd1R: Synthetic workloads based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

__kernel void mat3dadd(__global DATA_TYPE* A,
                       __global int* B,
                       __global DATA_TYPE* C,
                       DATA_TYPE alpha,
                       int NI,
                       int NJ,
                       int NK)
{
#if WORK_DIM == 3
                int i = get_global_id(0);
                int j = get_global_id(1);
                int k = get_global_id(2);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = A[B[k]];
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * A[B[k]];
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * A[B[k]];
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    A[B[k]];
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[B[k]];
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[B[k]];
                }
#endif
#elif WORK_DIM == 2
                int i = get_global_id(0);
                int j = get_global_id(1);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = A[B[k]];
                    }
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * A[B[k]];
                    }
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * A[B[k]];
                    }
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        A[B[k]];
                    }
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[B[k]];
                    }
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[B[k]];
                    }
                }
#endif
#elif WORK_DIM == 1
                int i = get_global_id(0);

#if COMPUTATION == 0
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = A[B[k]];
                        }
                    }
                }
#elif COMPUTATION == 1
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * A[B[k]];
                        }
                    }
                }
#elif COMPUTATION == 2
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * A[B[k]];
                        }
                    }
                }
#elif COMPUTATION == 4
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            A[B[k]];
                        }
                    }
                }
#elif COMPUTATION == 8
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[B[k]];
                        }
                    }
                }
#elif COMPUTATION == 16
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[B[k]];
                        }
                    }
                }
#endif
#endif
}


