// 1Mat3DAdd1T: Synthetic workloads based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable
#pragma OPENCL EXTENSION cl_khr_local_int32_extended_atomics : enable

__kernel void mat3dadd(__global DATA_TYPE* A,
                       __global DATA_TYPE* C,
                       DATA_TYPE alpha,
                       int NI,
                       int NJ,
                       int NK)
{
#if WORK_DIM == 3
                int i = get_global_id(0);
                int j = get_global_id(1);
                int k = get_global_id(2);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
#endif
#elif WORK_DIM == 2
                int i = get_global_id(0);
                int j = get_global_id(1);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
#endif
#elif WORK_DIM == 1
                int i = get_global_id(0);

#if COMPUTATION == 0
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#elif COMPUTATION == 1
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#elif COMPUTATION == 2
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#elif COMPUTATION == 4
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#elif COMPUTATION == 8
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#elif COMPUTATION == 16
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j];
                        }
                    }
                }
#endif
#endif
}

__kernel void mat3dadd_throttling(__global DATA_TYPE* A,
                                  __global DATA_TYPE* C,
                                  DATA_TYPE alpha,
                                  int NI,
                                  int NJ,
                                  int NK,
                                  int throttling_mod,
                                  int throttling_alloc)
{
    volatile __local int local_worklist[1];
    if (get_local_id(0) == 0)
        local_worklist[0] = 0;
    barrier(CLK_LOCAL_MEM_FENCE);

#if WORK_DIM == 3
    if (get_local_id(0) % throttling_mod < throttling_alloc) {
        for (int dynamic_work = atomic_inc(local_worklist);
                 dynamic_work < get_local_size(0) * get_local_size(1) * get_local_size(2);
                 dynamic_work = atomic_inc(local_worklist)) {
            int global_id_0 = get_group_id(0) * get_local_size(0) +
                              dynamic_work / (get_local_size(1) * get_local_size(2));
            int global_id_1 = get_group_id(1) * get_local_size(1) +
                              (dynamic_work % (get_local_size(1) * get_local_size(2))) / get_local_size(2);
            int global_id_2 = get_group_id(2) * get_local_size(2) +
                              dynamic_work % get_local_size(2);

            int i = global_id_0;
            int j = global_id_1;
            int k = global_id_2;

#if COMPUTATION == 0
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
            }
#elif COMPUTATION == 1
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                A[i * (NJ*NK) + k * (NK) + j];
            }
#elif COMPUTATION == 2
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                A[i * (NJ*NK) + k * (NK) + j];
            }
#elif COMPUTATION == 4
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                A[i * (NJ*NK) + k * (NK) + j];
            }
#elif COMPUTATION == 8
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                alpha * alpha * alpha * alpha *
                                                A[i * (NJ*NK) + k * (NK) + j];
            }
#elif COMPUTATION == 16
            if ((i < NI) && (j < NJ) && (k < NK)) {
                C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                alpha * alpha * alpha * alpha *
                                                alpha * alpha * alpha * alpha *
                                                alpha * alpha * alpha * alpha *
                                                A[i * (NJ*NK) + k * (NK) + j];
            }
#endif
        }
    }
#elif WORK_DIM == 2
    if (get_local_id(0) % throttling_mod < throttling_alloc) {
        for (int dynamic_work = atomic_inc(local_worklist);
                 dynamic_work < get_local_size(0) * get_local_size(1);
                 dynamic_work = atomic_inc(local_worklist)) {
            int global_id_0 = get_group_id(0) * get_local_size(0) +
                              dynamic_work / get_local_size(1);
            int global_id_1 = get_group_id(1) * get_local_size(1) +
                              dynamic_work % get_local_size(1);
            int i = global_id_0;
            int j = global_id_1;

#if COMPUTATION == 0
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#elif COMPUTATION == 1
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#elif COMPUTATION == 2
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#elif COMPUTATION == 4
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#elif COMPUTATION == 8
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#elif COMPUTATION == 16
            if ((i < NI) && (j < NJ)) {
                for (int k = 0; k < NK; k++) {
                    C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j];
                }
            }
#endif
        }
    }
#elif WORK_DIM == 1
    if (get_local_id(0) % throttling_mod < throttling_alloc) {
        for (int dynamic_work = atomic_inc(local_worklist);
                 dynamic_work < get_local_size(0);
                 dynamic_work = atomic_inc(local_worklist)) {
            int global_id_0 = get_group_id(0) * get_local_size(0) + dynamic_work;
            int i = global_id_0;

#if COMPUTATION == 0
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#elif COMPUTATION == 1
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#elif COMPUTATION == 2
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#elif COMPUTATION == 4
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#elif COMPUTATION == 8
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#elif COMPUTATION == 16
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j];
                    }
                }
            }
#endif
        }
    }
#endif
}

