#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

//// first index: total number of CPU cores of the system
//// second index: the number of arguments for this kernel
//unsigned long x86_kernel_params[44] = {
//    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
//    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
//    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
//};

unsigned long x86_kernel_params[100] = { 0, };

int num_total_cores = 4;
int num_total_params = 10;

/* Can switch DATA_TYPE between float and double */
#if defined(FLOAT)
typedef float DATA_TYPE;
#elif defined(DOUBLE)
typedef double DATA_TYPE;
#elif defined(INT)
typedef int DATA_TYPE;
#else
typedef int DATA_TYPE;
#endif

void mat3dadd(DATA_TYPE* A, DATA_TYPE* C, DATA_TYPE alpha,
              int NI, int NJ, int NK,
              size_t* global_size, size_t* local_size,
              size_t start_wg_id, size_t num_wgs)
{
#if WORK_DIM == 3
    int _group_id[3];
    int _global_id[3];

    for (size_t wg_id = start_wg_id; wg_id < start_wg_id+num_wgs; wg_id++) {
        _group_id[0] = wg_id / (global_size[1]/local_size[1] * global_size[2]/local_size[2]);
        _group_id[1] = (wg_id % (global_size[1]/local_size[1] * global_size[2]/local_size[2])) /
            (global_size[2] / local_size[2]);
        _group_id[2] = wg_id % (global_size[2] / local_size[2]);

        _global_id[0] = _group_id[0] * local_size[0];
        _global_id[1] = _group_id[1] * local_size[1];
        _global_id[2] = _group_id[2] * local_size[2];

        int global_id[3];
        for (global_id[0] = _global_id[0]; global_id[0] < _global_id[0] + local_size[0]; global_id[0]++) {
            for (global_id[1] = _global_id[1]; global_id[1] < _global_id[1] + local_size[1]; global_id[1]++) {
                for (global_id[2] = _global_id[2]; global_id[2] < _global_id[2] + local_size[2]; global_id[2]++) {
                    int i = global_id[0];
                    int j = global_id[1];
                    int k = global_id[2];

#if COMPUTATION == 0
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += A[i * (NJ*NK) + j * (NK) + k];
                    }
#elif COMPUTATION == 1
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
#elif COMPUTATION == 2
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += alpha * alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
#elif COMPUTATION == 4
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
#elif COMPUTATION == 8
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                            alpha * alpha * alpha * alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
#elif COMPUTATION == 16
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
#endif
                }
            }
        }
    }
#elif WORK_DIM == 2
    int _group_id[2];
    int _global_id[2];

    for (size_t wg_id = start_wg_id; wg_id < start_wg_id+num_wgs; wg_id++) {
        _group_id[0] = wg_id / (global_size[1] / local_size[1]);
        _group_id[1] = wg_id % (global_size[1] / local_size[1]);

        _global_id[0] = _group_id[0] * local_size[0];
        _global_id[1] = _group_id[1] * local_size[1];

        int global_id[2];
        for (global_id[0] = _global_id[0]; global_id[0] < _global_id[0] + local_size[0]; global_id[0]++) {
            for (global_id[1] = _global_id[1]; global_id[1] < _global_id[1] + local_size[1]; global_id[1]++) {
                int i = global_id[0]; // get_global_id(0);
                int j = global_id[1]; // get_global_id(0);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
#endif
            }
        }
    }
#elif WORK_DIM == 1
    for (size_t wg_id = start_wg_id; wg_id < start_wg_id + num_wgs; wg_id++) {
        for (size_t global_id = wg_id * local_size[0];
                global_id < wg_id * local_size[0] + local_size[0];
                global_id++) {
            int i = (int)global_id; // get_global_id(0);

#if COMPUTATION == 0
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#elif COMPUTATION == 1
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#elif COMPUTATION == 2
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#elif COMPUTATION == 4
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha * A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#elif COMPUTATION == 8
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#elif COMPUTATION == 16
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        C[i*NJ + j] += alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       alpha * alpha * alpha * alpha *
                                       A[i * (NJ*NK) + j * (NK) + k];
                    }
                }
            }
#endif
        }
    }
#endif
}

extern "C" {
int CPU_thread_entry(int id) {
    DATA_TYPE* A = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 0];
    DATA_TYPE* C = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 1];
    DATA_TYPE* alpha = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 2];
    int* NI = (int*)x86_kernel_params[num_total_params * id + 3];
    int* NJ = (int*)x86_kernel_params[num_total_params * id + 4];
    int* NK = (int*)x86_kernel_params[num_total_params * id + 5];
    size_t* global_size = (size_t*)x86_kernel_params[num_total_params * id + 6];
    size_t* local_size = (size_t*)x86_kernel_params[num_total_params * id + 7];
    size_t start_wg_id = (size_t)x86_kernel_params[num_total_params * id + 8];
    size_t num_wgs = (size_t)x86_kernel_params[num_total_params * id + 9];

    mat3dadd(A, C, *alpha, *NI, *NJ, *NK, global_size, local_size, start_wg_id, num_wgs);

    return 0;
}
}
