// 2Mat3DAddC1T: Synthetic workloads based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

#pragma OPENCL EXTENSION cl_amd_fp64 : enable
#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable
#pragma OPENCL EXTENSION cl_khr_global_int32_extended_atomics : enable

inline int gpu_first(__local int* tmp, __global atomic_int* worklist)
{
    if (get_local_id(2) == 0 && get_local_id(1) == 0 && get_local_id(0) == 0) {
        tmp[0] = atomic_fetch_add(worklist, 1);
    }
    barrier(CLK_LOCAL_MEM_FENCE);
    int current = tmp[0];
    return current;
}

inline int gpu_next(__local int* tmp, __global atomic_int* worklist)
{
    if (get_local_id(2) == 0 && get_local_id(1) == 0 && get_local_id(0) == 0) {
        tmp[0] = atomic_fetch_add(worklist, 1);
    }
    barrier(CLK_LOCAL_MEM_FENCE);
    int current = tmp[0];
    return current;
}

inline bool gpu_more(int current, int n_tasks)
{
    return (current < n_tasks);
}

__kernel void mat2dadd_throttling(__global DATA_TYPE* A,
                                            __global DATA_TYPE* B,
                                            __global DATA_TYPE* C,
                                            DATA_TYPE alpha,
                                            DATA_TYPE beta,
                                            int NI,
                                            int NJ,
                                            int throttling_mod,
                                            int throttling_alloc)
{
    volatile __local int local_worklist[1];
    if (get_local_id(0) == 0)
        local_worklist[0] = 0;
    barrier(CLK_LOCAL_MEM_FENCE);
#if WORK_DIM == 1
    if (get_local_id(0) % throttling_mod < throttling_alloc) {
        for (int dynamic_work = atomic_inc(local_worklist);
                 dynamic_work < get_local_size(0);
                 dynamic_work = atomic_inc(local_worklist)) {
            int global_id_0 = get_group_id(0) * get_local_size(0) + get_global_offset(0) + dynamic_work;
            int i = global_id_0;

#if COMPUTATION == 0
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                            C[i * (NJ) + j] = A[i * (NJ) + j] +
                                           B[i * (NJ) + j];
                    }
                }
#elif COMPUTATION == 1
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
												C[i * (NJ) + j] = alpha *
																			 A[i * (NJ) + j] +
																			 beta *
																			 B[i * (NJ) + j];
										}
                }
#elif COMPUTATION == 2
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
												C[i * (NJ) + j] = alpha * alpha *
																			 A[i * (NJ) + j] +
																			 beta * beta *
																			 B[i * (NJ) + j];
                    }
                }
#elif COMPUTATION == 4
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
												C[i * (NJ) + j] = alpha * alpha * alpha * alpha *
																			 A[i * (NJ) + j] +
																			 beta * beta * beta * beta *
																			 B[i * (NJ) + j];
                    }
                }
#elif COMPUTATION == 8
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
												C[i * (NJ) + j] = alpha * alpha * alpha * alpha *
																			 alpha * alpha * alpha * alpha *
																			 A[i * (NJ) + j] +
																			 beta * beta * beta * beta *
																			 beta * beta * beta * beta *
																			 B[i * (NJ) + j];
                    }
                }
#elif COMPUTATION == 16
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
												C[i * (NJ) + j] = alpha * alpha * alpha * alpha *
																			 alpha * alpha * alpha * alpha *
																			 alpha * alpha * alpha * alpha *
																			 alpha * alpha * alpha * alpha *
																			 A[i * (NJ) + j] +
																			 beta * beta * beta * beta *
																			 beta * beta * beta * beta *
																			 beta * beta * beta * beta *
																			 beta * beta * beta * beta *
																			 B[i * (NJ) + j];
										}
                }
#endif
        }
    }
#endif
}

