// 2Mat3DAdd1R1T: Synthetic workloads based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

__kernel void mat3dadd(__global DATA_TYPE* A,
                       __global DATA_TYPE* B,
                       __global int* C,
                       __global DATA_TYPE* D,
                       DATA_TYPE alpha,
                       DATA_TYPE beta,
                       int NI,
                       int NJ,
                       int NK)
{
#if WORK_DIM == 3
                int i = get_global_id(0);
                int j = get_global_id(1);
                int k = get_global_id(2);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j] +
                                                    B[C[k]];
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j] +
                                                    beta *
                                                    B[C[k]];
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j] +
                                                    beta * beta *
                                                    B[C[k]];
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j] +
                                                    beta * beta * beta * beta *
                                                    B[C[k]];
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j] +
                                                    beta * beta * beta * beta *
                                                    beta * beta * beta * beta *
                                                    B[C[k]];
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ) && (k < NK)) {
                    D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    alpha * alpha * alpha * alpha *
                                                    A[i * (NJ*NK) + k * (NK) + j] +
                                                    beta * beta * beta * beta *
                                                    beta * beta * beta * beta *
                                                    beta * beta * beta * beta *
                                                    beta * beta * beta * beta *
                                                    B[C[k]];
                }
#endif
#elif WORK_DIM == 2
                int i = get_global_id(0);
                int j = get_global_id(1);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j] +
                                                        B[C[k]];
                    }
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j] +
                                                        beta *
                                                        B[C[k]];
                    }
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j] +
                                                        beta * beta *
                                                        B[C[k]];
                    }
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j] +
                                                        beta * beta * beta * beta *
                                                        B[C[k]];
                    }
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j] +
                                                        beta * beta * beta * beta *
                                                        beta * beta * beta * beta *
                                                        B[C[k]];
                    }
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        alpha * alpha * alpha * alpha *
                                                        A[i * (NJ*NK) + k * (NK) + j] +
                                                        beta * beta * beta * beta *
                                                        beta * beta * beta * beta *
                                                        beta * beta * beta * beta *
                                                        beta * beta * beta * beta *
                                                        B[C[k]];
                    }
                }
#endif
#elif WORK_DIM == 1
                int i = get_global_id(0);

#if COMPUTATION == 0
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + k * (NK) + j] +
                                                            B[C[k]];
                        }
                    }
                }
#elif COMPUTATION == 1
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j] +
                                                            beta *
                                                            B[C[k]];
                        }
                    }
                }
#elif COMPUTATION == 2
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j] +
                                                            beta * beta *
                                                            B[C[k]];
                        }
                    }
                }
#elif COMPUTATION == 4
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j] +
                                                            beta * beta * beta * beta *
                                                            B[C[k]];
                        }
                    }
                }
#elif COMPUTATION == 8
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j] +
                                                            beta * beta * beta * beta *
                                                            beta * beta * beta * beta *
                                                            B[C[k]];
                        }
                    }
                }
#elif COMPUTATION == 16
                if (i < NI) {
                    for (int j = 0; j < NJ; j++) {
                        for (int k = 0; k < NK; k++) {
                            D[i * (NJ*NK) + j * (NK) + k] = alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            alpha * alpha * alpha * alpha *
                                                            A[i * (NJ*NK) + k * (NK) + j] +
                                                            beta * beta * beta * beta *
                                                            beta * beta * beta * beta *
                                                            beta * beta * beta * beta *
                                                            beta * beta * beta * beta *
                                                            B[C[k]];
                        }
                    }
                }
#endif
#endif
}


