#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iostream>

#include <atomic>

unsigned long x86_kernel_params[128] = { 0, };

int num_total_cores = 4;
int num_total_params = 14;

/* Can switch DATA_TYPE between float and double */
#if defined(FLOAT)
typedef float DATA_TYPE;
#elif defined(DOUBLE)
typedef double DATA_TYPE;
#elif defined(INT)
typedef int DATA_TYPE;
#else
typedef int DATA_TYPE;
#endif

inline int cpu_first(std::atomic_int* worklist)
{
    int current = worklist->fetch_add(1);
    return current;
}

inline int cpu_next(std::atomic_int* worklist)
{
    int current = worklist->fetch_add(1);
    return current;
}

inline bool cpu_more(int current, int n_tasks)
{
    return current < n_tasks;
}

void mat4dadd(DATA_TYPE* A, DATA_TYPE* B, int* C, DATA_TYPE* D,
        DATA_TYPE alpha, DATA_TYPE beta,
        int NI, int NJ, int NK, int NM,
        size_t* global_size, size_t* local_size,
        std::atomic_int* worklist, size_t num_tasks)
{
#if WORK_DIM == 3
    int _group_id[3];
    int _global_id[3];

    for (_group_id[0] = cpu_first(worklist); cpu_more(_group_id[0], num_tasks); _group_id[0] = cpu_next(worklist)) {
        for (_group_id[1] = 0; _group_id[1] < (global_size[1]/local_size[1]); _group_id[1]++) {
            for (_group_id[2] = 0; _group_id[2] < (global_size[2]/local_size[2]); _group_id[2]++) {
                _global_id[0] = _group_id[0] * local_size[0];
                _global_id[1] = _group_id[1] * local_size[1];
                _global_id[2] = _group_id[2] * local_size[2];

                int global_id[3];
                for (global_id[0] = _global_id[0]; global_id[0] < _global_id[0] + local_size[0]; global_id[0]++) {
                    for (global_id[1] = _global_id[1]; global_id[1] < _global_id[1] + local_size[1]; global_id[1]++) {
                        for (global_id[2] = _global_id[2]; global_id[2] < _global_id[2] + local_size[2]; global_id[2]++) {
                            int i = global_id[0];
                            int j = global_id[1];
                            int k = global_id[2];

#if COMPUTATION == 0
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                B[C[m]];
                        }
                    }
#elif COMPUTATION == 1
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * B[C[m]];
                        }
                    }
#elif COMPUTATION == 2
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta *
                                B[C[m]];
                        }
                    }
#elif COMPUTATION == 4
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
#elif COMPUTATION == 8
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
#elif COMPUTATION == 16
                    if ((i < NI) && (j < NJ) && (k < NK)) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
#endif
                        }
                    }
                }
            }
        }
    }
#elif WORK_DIM == 2
    int _group_id[2];
    int _global_id[2];

    for (_group_id[0] = cpu_first(worklist); cpu_more(_group_id[0], num_tasks); _group_id[0] = cpu_next(worklist)) {
        for (_group_id[1] = 0; _group_id[1] < (global_size[1]/local_size[1]); _group_id[1]++) {
            _global_id[0] = _group_id[0] * local_size[0];
            _global_id[1] = _group_id[1] * local_size[1];

            int global_id[2];
            for (global_id[0] = _global_id[0]; global_id[0] < _global_id[0] + local_size[0]; global_id[0]++) {
                for (global_id[1] = _global_id[1]; global_id[1] < _global_id[1] + local_size[1]; global_id[1]++) {
                    int i = global_id[0]; // get_global_id(0);
                    int j = global_id[1]; // get_global_id(0);

#if COMPUTATION == 0
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                B[C[m]];
                        }
                    }
                }
#elif COMPUTATION == 1
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * B[C[m]];
                        }
                    }
                }
#elif COMPUTATION == 2
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta *
                                B[C[m]];
                        }
                    }
                }
#elif COMPUTATION == 4
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
#elif COMPUTATION == 8
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
#elif COMPUTATION == 16
                if ((i < NI) && (j < NJ)) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
#endif
                }
            }
        }
    }
#elif WORK_DIM == 1
    for (int wg_id = cpu_first(worklist); cpu_more(wg_id, num_tasks); wg_id = cpu_next(worklist)) {
        for (size_t global_id = wg_id * local_size[0]; global_id < wg_id * local_size[0] + local_size[0]; global_id++) {
            int i = (int)global_id; // get_global_id(0);

#if COMPUTATION == 0
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                B[C[m]];
                        }
                    }
                }
            }
#elif COMPUTATION == 1
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            C[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * B[C[m]];
                        }
                    }
                }
            }
#elif COMPUTATION == 2
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta *
                                B[C[m]];
                        }
                    }
                }
            }
#elif COMPUTATION == 4
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
            }
#elif COMPUTATION == 8
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
            }
#elif COMPUTATION == 16
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
                    for (int k = 0; k < NK; k++) {
                        for (int m = 0; m < NM; m++) {
                            D[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] =
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                alpha * alpha * alpha * alpha *
                                A[i * (NJ*NK*NM) + j * (NK*NM) + k * (NM) + m] +
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                beta * beta * beta * beta *
                                B[C[m]];
                        }
                    }
                }
            }
#endif
        }
    }
#endif
}

extern "C" {
int CPU_thread_entry(int id) {
    DATA_TYPE* A = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 0];
    DATA_TYPE* B = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 1];
    int* C = (int*)x86_kernel_params[num_total_params * id + 2];
    DATA_TYPE* D = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 3];
    DATA_TYPE* alpha = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 4];
    DATA_TYPE* beta = (DATA_TYPE*)x86_kernel_params[num_total_params * id + 5];
    int* NI = (int*)x86_kernel_params[num_total_params * id + 6];
    int* NJ = (int*)x86_kernel_params[num_total_params * id + 7];
    int* NK = (int*)x86_kernel_params[num_total_params * id + 8];
    int* NM = (int*)x86_kernel_params[num_total_params * id + 9];
    size_t* global_size = (size_t*)x86_kernel_params[num_total_params * id + 10];
    size_t* local_size = (size_t*)x86_kernel_params[num_total_params * id + 11];
    std::atomic_int* worklist = (std::atomic_int*)x86_kernel_params[num_total_params * id + 12];
    size_t num_tasks = (size_t)x86_kernel_params[num_total_params * id + 13];

    mat4dadd(A, B, C, D, *alpha, *beta, *NI, *NJ, *NK, *NM, global_size, local_size, worklist, num_tasks);

    return 0;
}
}

