#!/usr/bin/env python

from os.path import exists
import os
import sys

import commands
import re
import multiprocessing

number_re = "+([0-9]+\.[01-9]+|[0-9]+)"

def get_values_arr(log, prefix):
    try:
        val = re.findall(prefix+number_re, log)
        return val
    except:
        return "-1.0"

def get_values(log, prefix):
    try:
        val = re.findall(prefix+number_re, log)[0]
        return val
    except:
        return "-1.0"

def get_values_idx(log, prefix, idx):
    try:
        val = re.findall(prefix+number_re, log)[idx]
        return val
    except:
        return "-1.0"

supported_datasize = {"d1":[16384,32768,65536],\
                      "d2":[128,256,512],\
                      "d3":[128,256,512]}

num_total_cores = {
    "skylake": 8,
    "kaveri": 4,
    "real": multiprocessing.cpu_count()
}

for arch in ["skylake", "kaveri"]:
    analysis = open("analysis."+arch+".csv", "w")
    analysis.write("workload,workdim,globalsize,localsize,")
    for opcode in ["read_write_constant", "read_write_continuous", "read_write_stride", "read_write_random", "arith_int","arith_float"]:
        analysis.write(opcode+",")
    analysis.write("solution_cpu,solution_gpu\n")

    for workload in ["1Mat3DAdd", "1Mat3DAdd1R", "1Mat3DAdd1T", "1Mat3DAddC", "1Mat3DAddC1R", "1Mat3DAddC1T", "2Mat3DAdd", "2Mat3DAdd1R", "2Mat3DAdd1T", "2Mat3DAdd1R1T", "2Mat3DAddC", "2Mat3DAddC1R", "2Mat3DAddC1T", "2Mat3DAddC1R1T", "1Mat4DAdd", "1Mat4DAdd1R", "1Mat4DAdd1T", "1Mat4DAddC", "1Mat4DAddC1R" ,"1Mat4DAddC1T", "2Mat4DAdd", "2Mat4DAdd1R", "2Mat4DAdd1T", "2Mat4DAdd1R1T", "2Mat4DAddC", "2Mat4DAddC1R", "2Mat4DAddC1T", "2Mat4DAddC1R1T"]:
        for data_type in ["f","i"]:
            for workdim in ["d1","d2"]:
                for computation in ["c0","c4","c16"]:
                    keyword_data_type = ""
                    keyword_workdim = ""
                    keyword_computation = ""

                    kernel_name = ""
                    if "3D" in workload:
                        kernel_name = "mat3dadd"
                    elif "4D" in workload:
                        kernel_name = "mat4dadd"

                    if data_type is "d":
                        keyword_data_type = " -D DOUBLE "
                    elif data_type is "f":
                        keyword_data_type = " -D FLOAT "
                    elif data_type is "i":
                        keyword_data_type = " -D INT "

                    if workdim is "d1":
                        if "4D" in workload:
                            keyword_workdim = " -D WORK_DIM=2 "
                        else:
                            keyword_workdim = " -D WORK_DIM=1 "
                    elif workdim is "d2":
                        keyword_workdim = " -D WORK_DIM=2 "
                    elif workdim is "d3":
                        keyword_workdim = " -D WORK_DIM=3 "

                    if computation is "c0":
                        keyword_computation = " -D COMPUTATION=0 "
                    elif computation is "c1":
                        keyword_computation = " -D COMPUTATION=1 "
                    elif computation is "c2":
                        keyword_computation = " -D COMPUTATION=2 "
                    elif computation is "c4":
                        keyword_computation = " -D COMPUTATION=4 "
                    elif computation is "c8":
                        keyword_computation = " -D COMPUTATION=8 "
                    elif computation is "c16":
                        keyword_computation = " -D COMPUTATION=16 "

                    compiler_command = "./scan " + \
                                       keyword_data_type + keyword_workdim + keyword_computation +\
                                       workload + "/" + kernel_name + "-kernel.cl"
                    print compiler_command
                    output = commands.getoutput(compiler_command)

                    features = []
                    value = 0
                    for opcode in ["read4_constant","write4_constant"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["read8_constant","write8_constant"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    features.append(value)

                    value = 0
                    for opcode in ["read4_continuous","write4_continuous"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["read8_continuous","write8_continuous"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    features.append(value)

                    value = 0
                    for opcode in ["read4_stride","write4_stride"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["read8_stride","write8_stride"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    features.append(value)

                    value = 0
                    for opcode in ["read4_random","write4_random"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["read8_random","write8_random"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    features.append(value)

                    value = 0
                    for opcode in ["add_int4", "add_int8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["mul_int4", "mul_int8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    for opcode in ["div_int4", "div_int8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 4*int(get_values(output, configuration + ": "))
                    features.append(value)

                    value = 0
                    for opcode in ["add_float4", "add_float8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += int(get_values(output, configuration + ": "))
                    for opcode in ["mul_float4", "mul_float8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 2*int(get_values(output, configuration + ": "))
                    for opcode in ["div_float4", "div_float8"]:
                        for loop_level in ["0","1","2"]:
                            configuration = kernel_name + "_loop" + loop_level + "_" + opcode
                            value += 4*int(get_values(output, configuration + ": "))
                    features.append(value)

                    for datasize in supported_datasize[workdim]:
                        for workconfig in [1,2]:
                            dim = 0
                            if workdim is "d1":
                                dim = 1
                            elif workdim is "d2":
                                dim = 2
                            elif workdim is "d3":
                                dim = 3

                            dataset_name = workload + "/" + "dataset-" + arch + "/" + kernel_name + "." + data_type + "." + \
                                           workdim + "." + computation + "." + str(datasize) + "." + str(dim) + "." + str(workconfig) + ".train.data"
                            if arch == "real":
                                dataset_name = workload + "/dataset/" + kernel_name + "." + data_type + "." + \
                                               workdim + "." + computation + "." + str(datasize) + "." + str(dim) + "." + str(workconfig) + ".train.data"
                            if exists(dataset_name):
                                dataset_f = open(dataset_name, "r")
                                data = dataset_f.readline()
                                numlines = num_total_cores[arch] * 8 + 8 + num_total_cores[arch]
                                dataset = []
                                for i in range(numlines):
                                    dataset.append(dataset_f.readline())

                                min_elapsed_time = 10000000000
                                for i in range(len(dataset)):
                                    elems = dataset[i].split(",")
                                    cpuresource = float(elems[3])/float(num_total_cores[arch])
                                    if cpuresource != 0.125 and cpuresource != 0.375 and cpuresource != 0.625 and cpuresource != 0.875:
                                        if min_elapsed_time > int(dataset[i].split(",")[5]):
                                            min_elapsed_time = int(dataset[i].split(",")[5])

                                for i in range(len(dataset)):
                                    elems = dataset[i].split(",")
                                    analysis.write(workload+"."+str(datasize)+"."+str(workconfig)+","+str(dim)+","+str(elems[1])+","+str(elems[2]))
                                    for j in range(len(features)):
                                        analysis.write("," + str(features[j]))
                                    analysis.write(","+str(int(float(elems[3])*num_total_cores[arch] + float(elems[4])*8))+"\n")

    analysis.close()

