#!/usr/bin/env python3

import sys
import os
import numpy as np
from multiprocessing import Pool
from sklearn import preprocessing
from sklearn.tree import DecisionTreeRegressor
from sklearn.externals import joblib

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print('Usage: ./ml.py CSV_INPUT_PATH');
        sys.exit(1)

    dataset_path = sys.argv[1]

    X_train = []
    Y_train = []
    with open(dataset_path, 'r') as fp:
        lines = fp.readlines()
        for line in lines[1:]:
            line = line.strip().split(',')
            X_train.append([float(v) for v in line[1:-1]])
            Y_train.append([float(line[-1])])

    X_train = np.array(X_train)
    Y_train = np.array(Y_train)

    X_scaler = preprocessing.StandardScaler()
    Y_scaler = preprocessing.StandardScaler()

    X_train = X_scaler.fit_transform(X_train)
    Y_train = Y_scaler.fit_transform(Y_train)

    regr = DecisionTreeRegressor()
    regr.fit(X_train, Y_train.ravel())
        
    model_path = os.getenv('ARMCL_ROOT') + '/model/'
    joblib.dump(regr, model_path + "model.pkl")
    joblib.dump(X_scaler, model_path + "X_scaler.pkl")
    joblib.dump(Y_scaler, model_path + "Y_scaler.pkl")
